#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import random
from .enums import Authentication, Security
from .GXDLMSChipperingStream import GXDLMSChipperingStream
from .GXByteBuffer import GXByteBuffer
from .CountType import CountType
from .GXDLMSChippering import GXDLMSChippering
#
class GXSecure:
    #
    #      * Chipher text.
    #      *
    #      * @param settings
    #      * DLMS settings.
    #      * @param cipher
    #      * Chipher settings.
    #      * @param ic
    #      * IC
    #      * @param data
    #      * Text to chipher.
    #      * @param secret
    #      * Secret.
    #      * @return Chiphered text.
    #pylint: disable=too-many-arguments
    @classmethod
    def secure(cls, settings, cipher, ic, data, secret):
        from .AesGcmParameter import AesGcmParameter
        if isinstance(secret, str):
            secret = secret.encode()
        d = []
        s = []
        if settings.authentication == Authentication.HIGH:
            len_ = len(data)
            if (len_ % 16) != 0:
                len_ += (16 - (len(data) % 16))

            if len(secret) > len(data):
                len_ = len(secret)
            if len_ % 16 != 0:
                len_ += 16 - (len(secret) % 16)
            s = bytearray(len_)
            d = bytearray(len_)
            s[0:len(secret)] = secret[0:]
            d[0:len(data)] = data[0:]
            pos = 0
            while pos < len(d) / 16:
                GXDLMSChipperingStream.aes1Encrypt(d, pos * 16, s)
                pos += 1
            return d
        #  Get server Challenge.
        challenge = GXByteBuffer()
        #  Get shared secret
        if settings.authentication == Authentication.HIGH_GMAC:
            challenge.set(data)
        else:
            challenge.set(data)
            challenge.set(secret)
        d = challenge.array()
        if settings.authentication == Authentication.HIGH_MD5:
            raise Exception("MD5is not supported.")
        if settings.authentication == Authentication.HIGH_SHA1:
            raise Exception("SHA1 is not supported.")
        if settings.authentication == Authentication.HIGH_SHA256:
            raise Exception("SHA256 is not supported.")
        if settings.authentication == Authentication.HIGH_GMAC:
            #  SC is always Security.Authentication.
            p = AesGcmParameter(0, secret, cipher.blockCipherKey, cipher.authenticationKey)
            p.security = Security.AUTHENTICATION
            p.invocationCounter = ic
            p.type_ = CountType.TAG
            challenge.clear()
            challenge.setUInt8(Security.AUTHENTICATION)
            challenge.setUInt32(p.invocationCounter)
            challenge.set(GXDLMSChippering.encryptAesGcm(p, d))
            d = challenge.array()
        if settings.authentication == Authentication.HIGH_ECDSA:
            raise Exception("ECDSA is not supported.")
        return d

    #
    #      * Generates challenge.
    #      *
    #      * @param authentication
    #      * Used authentication.
    #      * @return Generated challenge.
    #
    @classmethod
    def generateChallenge(cls):
        #  Random challenge is 8 to 64 bytes.
        #  Texas Instruments accepts only 16 byte long challenge.
        #  For this reason challenge size is 16 bytes at the moment.
        len_ = 16
        result = [None] * len_
        pos = 0
        while pos != len_:
            result[pos] = random.randint(0, 255)
            pos += 1
        return result



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#pylint: disable=broad-except,no-name-in-module
import time
from datetime import datetime
# If E0401: Unable to import 'dateutil.tz'
# pip install python-dateutil
from dateutil.tz import tzoffset
# If E0401: Unable to import 'pytz'
# pip install pytz
import pytz
from ._GXDataInfo import _GXDataInfo
from ..GXByteBuffer import GXByteBuffer
from ..GXBitString import GXBitString
from ..enums import DataType
from ..enums import DateTimeSkips, DateTimeExtraInfo, ClockStatus
from ..TranslatorTags import TranslatorTags
from ..TranslatorOutputType import TranslatorOutputType
from ..GXStructure import GXStructure
from ..GXArray import GXArray
from ..enums.Standard import Standard

# pylint: disable=too-many-public-methods
class _GXCommon:
    """This class is for internal use only and is subject to changes or removal
    in future versions of the API.  Don't use it."""

    #      HDLC frame start and end character.
    HDLC_FRAME_START_END = 0x7E
    LLC_SEND_BYTES = bytearray([0xE6, 0xE6, 0x00])
    LLC_REPLY_BYTES = bytearray([0xE6, 0xE7, 0x00])
    DATA_TYPE_OFFSET = 0xFF0000
    zeroes = "00000000000000000000000000000000"


    @classmethod
    def getBytes(cls, value):
        """
        Convert string to byte array.
        value: String value.
        returns String as bytes.
        """
        if not value:
            return None
        return value.encode()

    #
    # Is string hex string.
    #
    # value: String value.
    # Return true, if string is hex string.
    #
    @classmethod
    def isHexString(cls, value):
        # pylint: disable=chained-comparison
        if not value:
            return False
        ch = str()
        pos = 0
        while pos != len(value):
            ch = value.charAt(pos)
            if ch != ' ':
                if not ((ch > 0x40 and ch < 'G') or (ch > 0x60 and ch < 'g') or (ch > '/' and ch < ':')):
                    return False
            pos += 1
        return True

    #
    # Get object count.  If first byte is 0x80 or higger it will tell
    #      bytes
    # count.
    # data received data.
    # Object count.
    #
    @classmethod
    def getObjectCount(cls, data):
        cnt = data.getUInt8()
        if cnt > 0x80:
            if cnt == 0x81:
                cnt = data.getUInt8()
            elif cnt == 0x82:
                cnt = data.getUInt16()
            elif cnt == 0x84:
                cnt = int(data.getUInt32())
            else:
                raise ValueError("Invalid count.")
        return cnt

    #
    # Return how many bytes object count takes.
    #
    # count
    # Value
    # Value size in bytes.
    #
    @classmethod
    def getObjectCountSizeInBytes(cls, count):
        if count < 0x80:
            ret = 1
        elif count < 0x100:
            ret = 2
        elif count < 0x10000:
            ret = 3
        else:
            ret = 5
        return ret

    #
    # Add string to byte buffer.
    #
    # value
    # String to add.
    # bb
    # Byte buffer where string is added.
    #
    @classmethod
    def addString(cls, value, bb):
        bb.setUInt8(int(DataType.OCTET_STRING))
        if not value:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(value), bb)
            bb.set(value.encode())

    #
    # Set item count.
    # count
    # buff
    #
    @classmethod
    def setObjectCount(cls, count, buff):
        if count < 0x80:
            buff.setUInt8(count)
            ret = 1
        elif count < 0x100:
            buff.setUInt8(0x81)
            buff.setUInt8(count)
            ret = 2
        elif count < 0x10000:
            buff.setUInt8(0x82)
            buff.setUInt16(count)
            ret = 3
        else:
            buff.setUInt8(0x84)
            buff.setUInt32(count)
            ret = 5
        return ret

    #
    # Reserved for internal use.
    #
    @classmethod
    def toBitString(cls, value, count):
        count2 = count
        sb = ""
        if count2 > 0:
            if count2 > 8:
                count2 = 8
            pos = 7
            while pos != 8 - count2 - 1:
                if (value & (1 << pos)) != 0:
                    sb += '1'
                else:
                    sb += '0'
                pos -= 1
        return sb

    @classmethod
    def changeType(cls, value, type_):
        from ..GXDateTime import GXDateTime
        from ..GXDate import GXDate
        from ..GXTime import GXTime
        if value is None:
            ret = None
        elif type_ == DataType.NONE:
            ret = GXByteBuffer.hex(value, True)
        elif type_ in (DataType.STRING, DataType.OCTET_STRING) and not value:
            ret = ""
        elif type_ == DataType.DATETIME and not value:
            ret = GXDateTime(None)
        elif type_ == DataType.DATE and not value:
            ret = GXDate(None)
        elif type_ == DataType.TIME and not value:
            ret = GXTime(None)
        else:
            info = _GXDataInfo()
            info.type_ = type_
            ret = _GXCommon.getData(GXByteBuffer(value), info)
            if not info.complete:
                raise ValueError("Change type failed. Not enought data.")
            if type_ == DataType.OCTET_STRING and isinstance(ret, bytes):
                ret = GXByteBuffer.hex(ret)
        return ret

    #
    # Get data from DLMS frame.
    #
    # data
    # received data.
    # info
    # Data info.
    # Received data.
    #
    @classmethod
    def getData(cls, data, info):
        value = None
        startIndex = data.position
        if data.position == len(data):
            info.complete = False
            return None
        info.complete = True
        knownType = info.type_ != DataType.NONE
        #  Get data type if it is unknown.
        if not knownType:
            info.type_ = data.getUInt8()
        if info.type_ == DataType.NONE:
            if info.xml:
                info.xml.appendLine("<" + info.xml.getDataType(info.type_) + " />")
            return value
        if data.position == len(data):
            info.complete = False
            return None
        if info.type_ == DataType.ARRAY or info.type_ == DataType.STRUCTURE:
            value = cls.getArray(data, info, startIndex)
        elif info.type_ == DataType.BOOLEAN:
            value = cls.getBoolean(data, info)
        elif info.type_ == DataType.BITSTRING:
            value = cls.getBitString(data, info)
        elif info.type_ == DataType.INT32:
            value = cls.getInt32(data, info)
        elif info.type_ == DataType.UINT32:
            value = cls.getUInt32(data, info)
        elif info.type_ == DataType.STRING:
            value = cls.getString(data, info, knownType)
        elif info.type_ == DataType.STRING_UTF8:
            value = cls.getUtfString(data, info, knownType)
        elif info.type_ == DataType.OCTET_STRING:
            value = cls.getOctetString(data, info, knownType)
        elif info.type_ == DataType.BCD:
            value = cls.getBcd(data, info)
        elif info.type_ == DataType.INT8:
            value = cls.getInt8(data, info)
        elif info.type_ == DataType.INT16:
            value = cls.getInt16(data, info)
        elif info.type_ == DataType.UINT8:
            value = cls.getUInt8(data, info)
        elif info.type_ == DataType.UINT16:
            value = cls.getUInt16(data, info)
        elif info.type_ == DataType.COMPACT_ARRAY:
            value = cls.getCompactArray(data, info)
        elif info.type_ == DataType.INT64:
            value = cls.getInt64(data, info)
        elif info.type_ == DataType.UINT64:
            value = cls.getUInt64(data, info)
        elif info.type_ == DataType.ENUM:
            value = cls.getEnum(data, info)
        elif info.type_ == DataType.FLOAT32:
            value = cls.getFloat(data, info)
        elif info.type_ == DataType.FLOAT64:
            value = cls.getDouble(data, info)
        elif info.type_ == DataType.DATETIME:
            value = cls.getDateTime(data, info)
        elif info.type_ == DataType.DATE:
            value = cls.getDate(data, info)
        elif info.type_ == DataType.TIME:
            value = cls.getTime(data, info)
        else:
            raise ValueError("Invalid data type.")
        return value

    #
    # Convert value to hex string.
    # value value to convert.
    # desimals Amount of decimals.
    # @return
    #
    @classmethod
    def integerToHex(cls, value, desimals):
        str_ = hex(value)[2:]
        if desimals == 0 or len(_GXCommon.zeroes) == len(str_):
            return str_
        return _GXCommon.zeroes[0: desimals - len(str_)] + str_.upper()

    #
    # Convert value to hex string.
    # value value to convert.
    # desimals Amount of decimals.
    # @return
    #
    @classmethod
    def integerString(cls, value, desimals):
        str_ = str(value)
        if desimals == 0 or len(_GXCommon.zeroes) == len(str_):
            return str_
        return _GXCommon.zeroes[0: desimals - len(str_)] + str_

    #
    # Get array from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # index
    # starting index.
    # Object array.
    #
    @classmethod
    def getArray(cls, buff, info, index):
        value = None
        if info.count == 0:
            info.count = _GXCommon.getObjectCount(buff)
        if info.xml:
            info.xml.appendStartTag(info.xml.getDataType(info.type_), "Qty", info.xml.integerToHex(info.count, 2))
        size = len(buff) - buff.position
        if info.count != 0 and size < 1:
            info.complete = False
            return None
        startIndex = index
        if info.type_ == DataType.ARRAY:
            value = GXArray()
        else:
            value = GXStructure()
        #  Position where last row was found.  Cache uses this info.
        pos = info.index
        while pos != info.count:
            info2 = _GXDataInfo()
            info2.xml = info.xml
            tmp = cls.getData(buff, info2)
            if not info2.complete:
                buff.position = startIndex
                info.complete = False
                break
            else:
                if info2.count == info2.index:
                    startIndex = buff.position
                    value.append(tmp)
            pos += 1
        if info.xml:
            info.xml.appendEndTag(info.xml.getDataType(info.type_))
        info.index = pos
        return value

    #
    # Get time from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # Parsed time.
    #
    @classmethod
    def getTime(cls, buff, info):
        # pylint: disable=broad-except
        from ..GXTime import GXTime
        value = None
        if len(buff) - buff.position < 4:
            #  If there is not enough data available.
            info.complete = (False)
            return None
        str_ = None
        if info.xml:
            str_ = buff.toHex(False, buff.position, 4)
        try:
            value = GXTime(None)
            #  Get time.
            hour = buff.getUInt8()
            if hour == 0xFF:
                hour = 0
                value.skip |= DateTimeSkips.HOUR
            minute = buff.getUInt8()
            if minute == 0xFF:
                minute = 0
                value.skip |= DateTimeSkips.MINUTE
            second = buff.getUInt8()
            if second == 0xFF:
                second = 0
                value.skip |= DateTimeSkips.SECOND
            ms = buff.getUInt8()
            if ms != 0xFF:
                ms *= 10
            else:
                ms = 0
                value.skip |= DateTimeSkips.MILLISECOND
            value.value = datetime(1900, 1, 1, hour, minute, second, ms)
        except Exception as ex:
            if info.xml is None:
                raise ex
        if info.xml:
            if value:
                info.xml.appendComment(str(value))
            info.xml.appendLine(info.xml.getDataType(info.type_), None, str_)
        return value

    #
    # Get date from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # Parsed date.
    #
    @classmethod
    def getDate(cls, buff, info):
        # pylint: disable=broad-except
        from ..GXDate import GXDate
        value = None
        if len(buff) - buff.position < 5:
            #  If there is not enough data available.
            info.complete = (False)
            return None
        str_ = None
        if info.xml:
            str_ = buff.integerToHex(False, buff.position, 5)
        try:
            dt = GXDate()
            #  Get year.
            year = buff.getUInt16()
            if year < 1900 or year == 0xFFFF:
                dt.skip |= DateTimeSkips.YEAR
                year = 2000
            #  Get month
            month = buff.getUInt8()
            if month == 0xFE:
                dt.extra |= DateTimeExtraInfo.DST_BEGIN
                month = 1
            elif month != 0xFD:
                dt.extra |= DateTimeExtraInfo.DST_END
                month = 1
            else:
                if month < 1 or month > 12:
                    dt.skip |= DateTimeSkips.MONTH
                    month = 1
            #  Get day
            day = buff.getUInt8()
            if day == 0xFE:
                dt.extra |= DateTimeExtraInfo.LAST_DAY
                day = 1
            elif day != 0xFD:
                dt.extra |= DateTimeExtraInfo.LAST_DAY2
                day = 1
            else:
                if day < 1 or day > 31:
                    dt.skip |= DateTimeSkips.DAY
                    day = 1
            dt.value = datetime(year, month, day, 0, 0, 0, 0)
            value = dt
            #  Skip week day
            if buff.getUInt8() == 0xFF:
                dt.skip |= DateTimeSkips.DAY_OF_WEEK
        except Exception as ex:
            if info.xml is None:
                raise ex
        if info.xml:
            if value:
                info.xml.appendComment(str(value))
            info.xml.appendLine(info.xml.getDataType(info.type_), None, str_)
        return value

    #
    # Get date and time from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # Parsed date and time.
    #
    @classmethod
    def getDateTime(cls, buff, info):
        # pylint: disable=too-many-locals, broad-except
        from ..GXDateTime import GXDateTime
        value = None
        skip = DateTimeSkips.NONE
        extra = DateTimeExtraInfo.NONE
        #  If there is not enough data available.
        if len(buff) - buff.position < 12:
            info.complete = (False)
            return None
        str_ = None
        if info.xml:
            str_ = buff.toHex(False, buff.position, 12)
        dt = GXDateTime()
        try:
            #  Get year.
            year = buff.getUInt16()
            #  Get month
            month = buff.getUInt8()
            #  Get day
            day = buff.getUInt8()
            #  Skip week day
            dayOfWeek = buff.getUInt8()
            if dayOfWeek == 0xFF:
                skip |= DateTimeSkips.DAY_OF_WEEK
            else:
                dt.dayOfWeek = dayOfWeek
            #  Get time.
            hour = buff.getUInt8()
            minute = buff.getUInt8()
            second = buff.getUInt8()
            ms = buff.getUInt8() & 0xFF
            if ms != 0xFF:
                ms *= 10
            else:
                ms = -1
            deviation = buff.getInt16()
            if deviation == -32768:
                deviation = 0x8000
                skip |= DateTimeSkips.DEVITATION
            status = buff.getUInt8()
            dt.status = ClockStatus(status)
            if year < 1900 or year == 0xFFFF:
                skip |= DateTimeSkips.YEAR
                year = 2000
            if month == 0xFE:
                extra |= DateTimeExtraInfo.DST_BEGIN
                month = 1
            elif month != 0xFD:
                extra |= DateTimeExtraInfo.DST_END
                month = 1
            else:
                if month < 1 or month > 12:
                    skip |= DateTimeSkips.MONTH
                    month = 1

            if day == 0xFE:
                extra |= DateTimeExtraInfo.LAST_DAY
                day = 1
            elif day != 0xFD:
                extra |= DateTimeExtraInfo.LAST_DAY2
                day = 1
            else:
                if day == -1 or day == 0 or day > 31:
                    skip |= DateTimeSkips.DAY
                    day = 1

            if hour < 0 or hour > 24:
                skip |= DateTimeSkips.HOUR
                hour = 0
            if minute < 0 or minute > 60:
                skip |= DateTimeSkips.MINUTE
                minute = 0
            if second < 0 or second > 60:
                skip |= DateTimeSkips.SECOND
                second = 0
            #  If ms is Zero it's skipped.
            if ms < 0 or ms > 100:
                skip |= DateTimeSkips.MILLISECOND
                ms = 0
            tz = None
            if deviation != 0x8000:
                tz = -deviation

            if tz:
                dt.value = datetime(year, month, day, hour, minute, second, ms, tzinfo=tzoffset(None, tz))
            else:
                dt.value = datetime(year, month, day, hour, minute, second, ms)
            dt.skip = skip
            value = dt
        except Exception as ex:
            if info.xml is None:
                raise ex
        if info.xml:
            if dt.skip & DateTimeSkips.YEAR == 0 and value:
                info.xml.appendComment(str(value))
            info.xml.appendLine(info.xml.getDataType(info.type_), None, str_)
        return value

    #
    # Get double value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # Parsed double value.
    #
    @classmethod
    def getDouble(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 8:
            info.complete = (False)
            return None
        value = buff.getDouble()
        if info.xml:
            if info.xml.comments:
                info.xml.appendComment("{:.2f}".format(value))
            tmp = GXByteBuffer()
            cls.setData(tmp, DataType.FLOAT64, value)
            info.xml.appendLine(info.xml.getDataType(info.type_), None, GXByteBuffer.toHex(False, 1, len(tmp) - 1))
        return value

    #
    # Get float value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # Parsed float value.
    #
    @classmethod
    def getFloat(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 4:
            info.complete = (False)
            return None
        value = buff.getFloat()
        if info.xml:
            if info.xml.comments:
                info.xml.appendComment("{:.2f}".format(value))
            tmp = GXByteBuffer()
            cls.setData(tmp, DataType.FLOAT32, value)
            info.xml.appendLine(info.xml.getDataType(info.type_), None, tmp.toHex(False, 1, len(tmp) - 1))
        return value

    #
    # Get enumeration value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed enumeration value.
    #
    @classmethod
    def getEnum(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 1:
            info.complete = (False)
            return None
        value = buff.getUInt8()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 2))
        return value

    #
    # Get UInt64 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UInt64 value.
    #
    @classmethod
    def getUInt64(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 8:
            info.complete = (False)
            return None
        value = buff.getUInt64()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 16))
        return value

    #
    # Get Int64 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed Int64 value.
    #
    @classmethod
    def getInt64(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 8:
            info.complete = (False)
            return None
        value = buff.getInt64()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 16))
        return value

    #
    # Get UInt16 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UInt16 value.
    #
    @classmethod
    def getUInt16(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 2:
            info.complete = (False)
            return None
        value = buff.getUInt16()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 4))
        return value

    @classmethod
    def getCompactArrayItem(cls, buff, dt, list_, len_):
        if isinstance(dt, list):
            tmp2 = list()
            for it in dt:
                if isinstance(it, DataType):
                    cls.getCompactArrayItem(buff, it, tmp2, 1)
                else:
                    cls.getCompactArrayItem(buff, it, tmp2, 1)
            list_.append(tmp2)
            return

        tmp = _GXDataInfo()
        tmp.type_ = dt
        start = buff.position
        if dt == DataType.STRING:
            while buff.position - start < len_:
                tmp.clear()
                tmp.type = dt
                list_.append(cls.getString(buff, tmp, False))
                if not tmp.complete:
                    break
        elif dt == DataType.OCTET_STRING:
            while buff.position - start < len_:
                tmp.clear()
                tmp.type = dt
                list_.append(cls.getOctetString(buff, tmp, False))
                if not tmp.complete:
                    break
        else:
            while buff.position - start < len_:
                tmp.clear()
                tmp.type_ = dt
                list_.append(cls.getData(buff, tmp))
                if not tmp.complete:
                    break

    @classmethod
    def getDataTypes(cls, buff, cols, len_):
        dt = None
        pos = 0
        while pos != len_:
            dt = buff.getUInt8()
            if dt == DataType.ARRAY:
                cnt = buff.getUInt16()
                tmp = list()
                tmp2 = list()
                cls.getDataTypes(buff, tmp, 1)
                i = 0
                while i != cnt:
                    tmp2.append(tmp)
                    i += 1
                cols.append(tmp2)
            elif dt == DataType.STRUCTURE:
                tmp = list()
                cls.getDataTypes(buff, tmp, buff.getUInt8())
                cols.append(tmp)
            else:
                cols.append(dt)
            pos += 1

    @classmethod
    def appendDataTypeAsXml(cls, cols, info):
        for it in cols:
            if isinstance(it, (DataType,)):
                info.xml.appendEmptyTag(info.xml.getDataType(it))
            elif isinstance(it, GXStructure):
                info.xml.appendStartTag(cls.DATA_TYPE_OFFSET + DataType.STRUCTURE, None, None)
                cls.appendDataTypeAsXml(it, info)
                info.xml.appendEndTag(cls.DATA_TYPE_OFFSET + DataType.STRUCTURE)
            elif isinstance(it, GXArray):
                info.xml.appendStartTag(cls.DATA_TYPE_OFFSET + DataType.ARRAY, None, None)
                cls.appendDataTypeAsXml(it, info)
                info.xml.appendEndTag(cls.DATA_TYPE_OFFSET + DataType.ARRAY)

    #
    # Get compact array value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UInt16 value.
    #
    @classmethod
    def getCompactArray(cls, buff, info):
        # pylint: disable=too-many-nested-blocks

        #  If there is not enough data available.
        if len(buff) - buff.position < 2:
            info.complete = (False)
            return None
        dt = buff.getUInt8()
        if dt == DataType.ARRAY:
            raise ValueError("Invalid compact array data.")
        len_ = _GXCommon.getObjectCount(buff)
        list_ = list()
        if dt == DataType.STRUCTURE:
            #  Get data types.
            cols = list()
            cls.getDataTypes(buff, cols, len_)
            len_ = _GXCommon.getObjectCount(buff)
            if info.xml:
                info.xml.appendStartTag(info.xml.getDataType(DataType.COMPACT_ARRAY), None, None)
                info.xml.appendStartTag(TranslatorTags.CONTENTS_DESCRIPTION)
                cls.appendDataTypeAsXml(cols, info)
                info.xml.appendEndTag(TranslatorTags.CONTENTS_DESCRIPTION)
                if info.xml.outputType == TranslatorOutputType.STANDARD_XML:
                    info.xml.appendStartTag(TranslatorTags.ARRAY_CONTENTS, None, None, True)
                    info.xml.append(buff.remainingHexString(True))
                    info.xml.appendEndTag(TranslatorTags.ARRAY_CONTENTS, True)
                    info.xml.appendEndTag(info.xml.getDataType(DataType.COMPACT_ARRAY))
                else:
                    info.xml.appendStartTag(TranslatorTags.ARRAY_CONTENTS)
            start = buff.position
            while buff.position - start < len_:
                row = list()
                pos = 0
                while pos != len(cols):
                    if isinstance(cols[pos], GXArray):
                        cls.getCompactArrayItem(buff, cols[pos], row, 1)
                    elif isinstance(cols[pos], GXStructure):
                        tmp2 = list()
                        cls.getCompactArrayItem(buff, cols[pos], tmp2, 1)
                        row.append(tmp2[0])
                    else:
                        cls.getCompactArrayItem(buff, cols[pos], row, 1)
                    if buff.position == len(buff):
                        break
                    pos += 1
                #  If all columns are read.
                if len(row) >= len(cols):
                    list_.append(row)
                else:
                    break
            if info.xml and info.xml.outputType == TranslatorOutputType.SIMPLE_XML:
                sb = ""
                for row in list_:
                    for it in row:
                        if isinstance(it, bytearray):
                            sb += GXByteBuffer.hex(it)
                        elif isinstance(it, list):
                            start = len(sb)
                            for it2 in it:
                                if isinstance(it2, bytearray):
                                    sb += GXByteBuffer.hex(it2)
                                else:
                                    sb += str(it2)
                                sb += ";"
                            if start != len(sb):
                                sb = sb[0:len(sb) - 1]
                        else:
                            sb += str(it)
                        sb += ";"
                    if sb:
                        sb = sb[0:len(sb) - 1]
                    info.xml.appendLine(sb)
                    sb = ""
            if info.xml and info.xml.outputType == TranslatorOutputType.SIMPLE_XML:
                info.xml.appendEndTag(TranslatorTags.ARRAY_CONTENTS)
                info.xml.appendEndTag(info.xml.getDataType(DataType.COMPACT_ARRAY))
        else:
            if info.xml:
                info.xml.appendStartTag(info.xml.getDataType(DataType.COMPACT_ARRAY), None, None)
                info.xml.appendStartTag(TranslatorTags.CONTENTS_DESCRIPTION)
                info.xml.appendEmptyTag(info.xml.getDataType(dt))
                info.xml.appendEndTag(TranslatorTags.CONTENTS_DESCRIPTION)
                info.xml.appendStartTag(TranslatorTags.ARRAY_CONTENTS, None, None, True)
                if info.xml.outputType == TranslatorOutputType.STANDARD_XML:
                    info.xml.append(buff.remainingHexString(False))
                    info.xml.appendEndTag(TranslatorTags.ARRAY_CONTENTS, True)
                    info.xml.appendEndTag(info.xml.getDataType(DataType.COMPACT_ARRAY))
            cls.getCompactArrayItem(buff, dt, list_, len_)
            if info.xml and info.xml.outputType == TranslatorOutputType.SIMPLE_XML:
                for it in list_:
                    if isinstance(it, bytearray):
                        info.xml.append(GXByteBuffer.hex(it))
                    else:
                        info.xml.append(str(it))
                    info.xml.append(";")
                if list_:
                    info.xml.setXmlLength(info.xml.getXmlLength() - 1)
                info.xml.appendEndTag(TranslatorTags.ARRAY_CONTENTS, True)
                info.xml.appendEndTag(info.xml.getDataType(DataType.COMPACT_ARRAY))
        return list_

    #
    # Get UInt8 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UInt8 value.
    #
    @classmethod
    def getUInt8(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 1:
            info.complete = (False)
            return None
        value = buff.getUInt8() & 0xFF
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 2))
        return value

    #
    # Get Int16 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed Int16 value.
    #
    @classmethod
    def getInt16(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 2:
            info.complete = (False)
            return None
        value = int(buff.getInt16())
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 4))
        return value

    #
    # Get Int8 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed Int8 value.
    #
    @classmethod
    def getInt8(cls, buff, info):
        value = None
        #  If there is not enough data available.
        if len(buff) - buff.position < 1:
            info.complete = (False)
            return None
        value = int(buff.getInt8())
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 2))
        return value

    #
    # Get BCD value from DLMS data.
    #
    # buff: Received DLMS data.
    # info: Data info.
    # Returns parsed BCD value.
    #
    @classmethod
    def getBcd(cls, buff, info):
        #  If there is not enough data available.
        if len(buff) - buff.position < 1:
            info.complete = (False)
            return None
        value = buff.getUInt8()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 2))
        return value

    #
    # Get UTF string value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UTF string value.
    #
    @classmethod
    def getUtfString(cls, buff, info, knownType):
        if knownType:
            len_ = len(buff)
        else:
            len_ = _GXCommon.getObjectCount(buff)
            #  If there is not enough data available.
            if len(buff) - buff.position < len_:
                info.complete = (False)
                return None
        if len_ > 0:
            value = buff.getString(buff.position, len_)
        else:
            value = ""
        if info.xml:
            if info.xml.getShowStringAsHex:
                info.xml.appendLine(info.xml.getDataType(info.type_), None, buff.toHex(False, buff.position - len, len))
            else:
                info.xml.appendLine(info.xml.getDataType(info.type_), None, value)
        return value

    #
    # Get octet string value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed octet string value.
    #
    @classmethod
    def getOctetString(cls, buff, info, knownType):
        # pylint: disable=too-many-nested-blocks,broad-except
        value = None
        if knownType:
            len_ = len(buff)
        else:
            len_ = _GXCommon.getObjectCount(buff)
            #  If there is not enough data available.
            if len(buff) - buff.position < len_:
                info.complete = (False)
                return None
        tmp = bytearray(len_)
        buff.get(tmp)
        value = tmp
        if info.xml:
            if info.xml.comments and tmp:
                #  This might be logical name.
                if len(tmp) == 6 and tmp[5] == -1:
                    info.xml.appendComment(cls.toLogicalName(tmp))
                else:
                    isString = True
                    #  Try to move octect string to DateTie, Date or time.
                    if len(tmp) == 5 or len(tmp) == 5 or len(tmp) == 4:
                        try:
                            type_ = None
                            if len(tmp) == 12:
                                type_ = DataType.DATETIME
                            elif len(tmp) == 5:
                                type_ = DataType.DATE
                            else:
                                type_ = DataType.TIME
                            dt = _GXCommon.changeType(tmp, type_)
                            year = dt.value.year
                            if 1970 < year > 2100:
                                info.xml.appendComment(str(dt))
                                isString = False
                        except Exception:
                            isString = True
                    if isString:
                        for it in tmp:
                            if it < 32 or it > 126:
                                isString = False
                                break
                        if isString:
                            info.xml.appendComment(str(tmp))
            info.xml.appendLine(info.xml.getDataType(info.type_), None, GXByteBuffer.hex(tmp, False))
        return value

    #
    # Get string value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed string value.
    #
    @classmethod
    def getString(cls, buff, info, knownType):
        value = None
        if knownType:
            len_ = len(buff)
        else:
            len_ = _GXCommon.getObjectCount(buff)
            #  If there is not enough data available.
            if len(buff) - buff.position < len_:
                info.complete = (False)
                return None
        if len_ > 0:
            value = buff.getString(buff.position, len_)
            buff.position += len_
        else:
            value = ""
        if info.xml:
            if info.xml.showStringAsHex:
                info.xml.appendLine(info.xml.getDataType(info.type_), None, buff.toHex(False, buff.position - len_, len_))
            else:
                info.xml.appendLine(info.xml.getDataType(info.type_), None, value)
        return value

    #
    # Get UInt32 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed UInt32 value.
    #
    @classmethod
    def getUInt32(cls, buff, info):
        #  If there is not enough data available.
        if len(buff) - buff.position < 4:
            info.complete = False
            return None
        value = buff.getUInt32()
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 8))
        return value

    #
    # Get Int32 value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed Int32 value.
    #
    @classmethod
    def getInt32(cls, buff, info):
        #  If there is not enough data available.
        if len(buff) - buff.position < 4:
            info.complete = (False)
            return None
        value = int(buff.getInt32())
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, info.xml.integerToHex(value, 8))
        return value

    #
    # Get bit string value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed bit string value.
    #
    @classmethod
    def getBitString(cls, buff, info):
        cnt = cls.getObjectCount(buff)
        t = cnt
        t /= 8
        if cnt % 8 != 0:
            t += 1
        byteCnt = int(t)
        #  If there is not enough data available.
        if len(buff) - buff.position < byteCnt:
            info.complete = (False)
            return None
        sb = ""
        while cnt > 0:
            sb += cls.toBitString(buff.getInt8(), cnt)
            cnt -= 8
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, sb)
        return GXBitString(sb)

    #
    # Get boolean value from DLMS data.
    #
    # buff
    # Received DLMS data.
    # info
    # Data info.
    # parsed boolean value.
    #
    @classmethod
    def getBoolean(cls, buff, info):
        #  If there is not enough data available.
        if len(buff) - buff.position < 1:
            info.complete = (False)
            return None
        value = bool(buff.getUInt8() != 0)
        if info.xml:
            info.xml.appendLine(info.xml.getDataType(info.type_), None, value.__str__())
        return value

    #
    # Get HDLC address from byte array.
    #
    # buff
    # byte array.
    # HDLC address.
    #
    @classmethod
    def getHDLCAddress(cls, buff):
        size = 0
        pos = buff.position
        while pos != len(buff):
            size += 1
            if buff.getUInt8(pos) & 0x1 == 1:
                break
            pos += 1
        if size == 1:
            ret = (buff.getUInt8() & 0xFE) >> 1
        elif size == 2:
            ret = buff.getUInt16()
            ret = ((ret & 0xFE) >> 1) | ((ret & 0xFE00) >> 2)
        elif size == 4:
            ret = buff.getUInt32()
            ret = ((ret & 0xFE) >> 1) | ((ret & 0xFE00) >> 2) | ((ret & 0xFE0000) >> 3) | ((ret & 0xFE000000) >> 4)
        else:
            raise ValueError("Wrong size.")
        return ret

    #
    # Convert object to DLMS bytes.
    #
    # buff
    # Byte buffer where data is write.
    # dataType
    # Data type.
    # value
    # Added Value.
    #
    @classmethod
    def setData(cls, buff, dataType, value):
        from ..GXDateTime import GXDateTime
        from ..GXDate import GXDate
        from ..GXTime import GXTime
        if dataType in (DataType.ARRAY, DataType.STRUCTURE) and isinstance(value, (GXByteBuffer, bytearray, bytes)):
            #  If byte array is added do not add type.
            buff.set(value)
            return

        buff.setUInt8(dataType)
        if dataType == DataType.NONE:
            pass
        elif dataType == DataType.BOOLEAN:
            if value:
                buff.setUInt8(1)
            else:
                buff.setUInt8(0)
        elif dataType == DataType.UINT8:
            buff.setUInt8(value)
        elif dataType in (DataType.INT8, DataType.ENUM):
            buff.setInt8(value)
        elif dataType in (DataType.UINT16, DataType.INT16):
            buff.setUInt16(value)
        elif dataType in (DataType.UINT32, DataType.INT32):
            buff.setUInt32(value)
        elif dataType in (DataType.UINT64, DataType.INT64):
            buff.setUInt64(value)
        elif dataType == DataType.FLOAT32:
            buff.setFloat(value)
        elif dataType == DataType.FLOAT64:
            buff.setDouble(value)
        elif dataType == DataType.BITSTRING:
            cls.setBitString(buff, value, True)
        elif dataType == DataType.STRING:
            cls.setString(buff, value)
        elif dataType == DataType.STRING_UTF8:
            cls.setUtfString(buff, value)
        elif dataType == DataType.OCTET_STRING:
            if isinstance(value, GXDate):
                #  Add size
                buff.setUInt8(5)
                cls.setDate(buff, value)
            elif isinstance(value, GXTime):
                #  Add size
                buff.setUInt8(4)
                cls.setTime(buff, value)
            elif isinstance(value, (GXDateTime, datetime)):
                buff.setUInt8(12)
                cls.setDateTime(buff, value)
            else:
                cls.setOctetString(buff, value)
        elif dataType in (DataType.ARRAY, DataType.STRUCTURE):
            cls.setArray(buff, value)
        elif dataType == DataType.BCD:
            cls.setBcd(buff, value)
        elif dataType == DataType.COMPACT_ARRAY:
            #  Compact array is not work with python because we don't know data
            #  types of each element.
            raise ValueError("Invalid data type.")
        elif dataType == DataType.DATETIME:
            cls.setDateTime(buff, value)
        elif dataType == DataType.DATE:
            cls.setDate(buff, value)
        elif dataType == DataType.TIME:
            cls.setTime(buff, value)
        else:
            raise ValueError("Invalid data type.")

    #
    # Convert time to DLMS bytes.
    #
    # buff
    # Byte buffer where data is write.
    # value
    # Added value.
    #
    @classmethod
    def setTime(cls, buff, value):
        dt = cls.__getDateTime(value)
        #  Add time.
        if dt.skip & DateTimeSkips.HOUR != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.hour)
        if dt.skip & DateTimeSkips.MINUTE != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.minute)
        if dt.skip & DateTimeSkips.SECOND != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.second)
        if dt.skip & DateTimeSkips.MILLISECOND != DateTimeSkips.NONE:
            #  Hundredth of seconds is not used.
            buff.setUInt8(0xFF)
        else:
            ms = dt.value.microsecond
            if ms != 0:
                ms /= 10000
            buff.setUInt8(int(ms))

    #
    # Convert date to DLMS bytes.
    #
    # buff
    # Byte buffer where data is write.
    # value
    # Added value.
    #
    @classmethod
    def setDate(cls, buff, value):
        dt = cls.__getDateTime(value)
        #  Add year.
        if dt.skip & DateTimeSkips.YEAR != DateTimeSkips.NONE:
            buff.setUInt16(0xFFFF)
        else:
            buff.setUInt16(dt.value.year)
        #  Add month
        if dt.extra & DateTimeExtraInfo.DST_BEGIN != 0:
            buff.setUInt8(0xFD)
        elif dt.extra & DateTimeExtraInfo.DST_END != 0:
            buff.setUInt8(0xFE)
        elif dt.skip & DateTimeSkips.MONTH != 0:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.month)
        #  Add day
        if dt.extra & DateTimeExtraInfo.LAST_DAY2 != DateTimeSkips.NONE:
            buff.setUInt8(0xFD)
        elif dt.extra & DateTimeExtraInfo.LAST_DAY != DateTimeSkips.NONE:
            buff.setUInt8(0xFE)
        elif dt.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.day)

        #  Day of week.
        if dt.skip & DateTimeSkips.DAY_OF_WEEK != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            if dt.dayOfWeek == 0:
                buff.setUInt8(dt.value.weekday() + 1)
            else:
                buff.setUInt8(dt.dayOfWeek)

    @classmethod
    def __getDateTime(cls, value):
        from ..GXDateTime import GXDateTime
        dt = None
        if isinstance(value, (GXDateTime)):
            dt = value
        elif isinstance(value, (datetime, str)):
            dt = GXDateTime(value)
            dt.skip |= DateTimeSkips.MILLISECOND
        else:
            raise ValueError("Invalid date format.")
        return dt

    #
    # Convert date time to DLMS bytes.
    #
    # buff
    # Byte buffer where data is write.
    # value
    # Added value.
    #
    @classmethod
    def setDateTime(cls, buff, value):
        dt = cls.__getDateTime(value)
        #  Add year.
        if dt.skip & DateTimeSkips.YEAR != DateTimeSkips.NONE:
            buff.setUInt16(0xFFFF)
        else:
            buff.setUInt16(dt.value.year)
        #  Add month
        if dt.extra & DateTimeExtraInfo.DST_BEGIN != 0:
            buff.setUInt8(0xFD)
        elif dt.extra & DateTimeExtraInfo.DST_END != 0:
            buff.setUInt8(0xFE)
        elif dt.skip & DateTimeSkips.MONTH != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.month)

        #  Add day
        if dt.extra & DateTimeExtraInfo.LAST_DAY2 != DateTimeSkips.NONE:
            buff.setUInt8(0xFD)
        elif dt.extra & DateTimeExtraInfo.LAST_DAY != DateTimeSkips.NONE:
            buff.setUInt8(0xFE)
        elif dt.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.day)
        #  Day of week.
        if dt.skip & DateTimeSkips.DAY_OF_WEEK != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            if dt.dayOfWeek == 0:
                buff.setUInt8(dt.value.weekday() + 1)
            else:
                buff.setUInt8(dt.dayOfWeek)
        #  Add time.
        if dt.skip & DateTimeSkips.HOUR != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.hour)
        if dt.skip & DateTimeSkips.MINUTE != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.minute)
        if dt.skip & DateTimeSkips.SECOND != DateTimeSkips.NONE:
            buff.setUInt8(0xFF)
        else:
            buff.setUInt8(dt.value.second)
        if dt.skip & DateTimeSkips.MILLISECOND != DateTimeSkips.NONE:
            #  Hundredth of seconds is not used.
            buff.setUInt8(0xFF)
        else:
            ms = dt.value.microsecond
            if ms != 0:
                ms /= 10000
            buff.setUInt8(int(ms))
        #  devitation not used.
        if dt.skip & DateTimeSkips.DEVITATION != DateTimeSkips.NONE:
            buff.setUInt16(0x8000)
        else:
            #  Add devitation.
            if dt.value.tzinfo:
                val = -dt.value.utcoffset().total_seconds()
                if dt.value.dst():
                    val -= 60
                buff.setUInt16(int(val))
            else:
                buff.setUInt16(int(time.timezone / 60))

        #  Add clock_status
        if dt.skip & DateTimeSkips.STATUS == DateTimeSkips.NONE:
            if dt.value.dst() or dt.status & ClockStatus.DAYLIGHT_SAVE_ACTIVE != ClockStatus.OK:
                buff.setUInt8(dt.status | ClockStatus.DAYLIGHT_SAVE_ACTIVE)
            else:
                buff.setUInt8(dt.status)
        else:
            buff.setUInt8(0xFF)

    @classmethod
    def setBcd(cls, buff, value):
        buff.setUInt8(value)

    @classmethod
    def setArray(cls, buff, value):
        if value:
            _GXCommon.setObjectCount(len(value), buff)
            for it in value:
                cls.setData(buff, cls.getDLMSDataType(it), it)
        else:
            _GXCommon.setObjectCount(0, buff)

    @classmethod
    def setOctetString(cls, buff, value):
        if isinstance(value, str):
            tmp = GXByteBuffer.hexToBytes(value)
            _GXCommon.setObjectCount(len(tmp), buff)
            buff.set(tmp)
        elif isinstance(value, GXByteBuffer):
            cls.setObjectCount(len(value), buff)
            buff.set(value)
        elif isinstance(value, (bytearray, bytes)):
            cls.setObjectCount(len(value), buff)
            buff.set(value)
        elif value is None:
            cls.setObjectCount(0, buff)
        else:
            raise ValueError("Invalid data type.")

    @classmethod
    def setUtfString(cls, buff, value):
        if value:
            tmp = value.encode()
            _GXCommon.setObjectCount(len(tmp), buff)
            buff.set(tmp)
        else:
            buff.setUInt8(0)

    @classmethod
    def setString(cls, buff, value):
        if value:
            _GXCommon.setObjectCount(len(value), buff)
            buff.set(_GXCommon.getBytes(value))
        else:
            buff.setUInt8(0)

    @classmethod
    def setBitString(cls, buff, val1, addCount):
        value = val1
        if isinstance(value, GXBitString):
            value = value.value
        if isinstance(value, str):
            val = 0
            str_ = str(value)
            if addCount:
                _GXCommon.setObjectCount(len(str_), buff)
            index = 7
            pos = 0
            while pos != len(str_):
                it = str_[pos]
                if it == '1':
                    val |= (1 << index)
                elif it != '0':
                    raise ValueError("Not a bit string.")
                index -= 1
                if index == -1:
                    index = 7
                    buff.setUInt8(val)
                    val = 0
                pos += 1
            if index != 7:
                buff.setUInt8(val)
        elif isinstance(value, (bytearray, bytes)):
            _GXCommon.setObjectCount(8 * len(value), buff)
            buff.set(value)
        elif isinstance(value, int):
            _GXCommon.setObjectCount(8, buff)
            buff.setUInt8(value)
        elif value is None:
            buff.setUInt8(0)
        else:
            raise ValueError("BitString must give as string.")

    @classmethod
    def getDataType(cls, value):
        from ..GXDateTime import GXDateTime
        from ..GXDate import GXDate
        from ..GXTime import GXTime
        if value == DataType.NONE:
            ret = None
        elif value == DataType.OCTET_STRING:
            ret = bytes.__class__
        elif value == DataType.ENUM:
            ret = int.__class__
            #ret = Enum.__class__
        elif value == DataType.INT8:
            ret = int.__class__
        elif value == DataType.INT16:
            ret = int.__class__
        elif value == DataType.INT32:
            ret = int.__class__
        elif value == DataType.INT64:
            ret = int.__class__
        elif value == DataType.TIME:
            ret = GXTime.__class__
        elif value == DataType.DATE:
            ret = GXDate.__class__
        elif value == DataType.DATETIME:
            ret = GXDateTime.__class__
        elif value == DataType.ARRAY:
            ret = list.__class__
        elif value == DataType.STRING:
            ret = str.__class__
        elif value == DataType.BOOLEAN:
            ret = bool.__class__
        elif value == DataType.FLOAT32:
            ret = float.__class__
        elif value == DataType.FLOAT64:
            ret = float.__class__
        else:
            raise ValueError("Invalid value.")
        return ret

    @classmethod
    def getDLMSDataType(cls, value):
        from ..GXDateTime import GXDateTime
        from ..GXDate import GXDate
        from ..GXTime import GXTime
        if value is None:
            ret = DataType.NONE
        elif isinstance(value, (bytes, bytearray, GXByteBuffer)):
            ret = DataType.OCTET_STRING
        #elif isinstance(value, (Enum,)):
        #    ret = DataType.ENUM
        elif isinstance(value, int):
            ret = DataType.INT32
        elif isinstance(value, GXTime):
            ret = DataType.TIME
        elif isinstance(value, GXDate):
            ret = DataType.DATE
        elif isinstance(value, (datetime, GXDateTime)):
            ret = DataType.DATETIME
        elif isinstance(value, list):
            ret = DataType.ARRAY
        elif isinstance(value, str):
            ret = DataType.STRING
        elif isinstance(value, bool):
            ret = DataType.BOOLEAN
        elif isinstance(value, float):
            ret = DataType.FLOAT32
        else:
            raise ValueError("Invalid value.")
        return ret

    @classmethod
    def getDataTypeSize(cls, type_):
        size = -1
        if type_ == DataType.BCD:
            size = 1
        elif type_ == DataType.BOOLEAN:
            size = 1
        elif type_ == DataType.DATE:
            size = 5
        elif type_ == DataType.DATETIME:
            size = 12
        elif type_ == DataType.ENUM:
            size = 1
        elif type_ == DataType.FLOAT32:
            size = 4
        elif type_ == DataType.FLOAT64:
            size = 8
        elif type_ == DataType.INT16:
            size = 2
        elif type_ == DataType.INT32:
            size = 4
        elif type_ == DataType.INT64:
            size = 8
        elif type_ == DataType.INT8:
            size = 1
        elif type_ == DataType.NONE:
            size = 0
        elif type_ == DataType.TIME:
            size = 4
        elif type_ == DataType.UINT16:
            size = 2
        elif type_ == DataType.UINT32:
            size = 4
        elif type_ == DataType.UINT64:
            size = 8
        elif type_ == DataType.UINT8:
            size = 1
        return size

    @classmethod
    def toLogicalName(cls, value):
        if isinstance(value, bytearray):
            if not value:
                value = bytearray(6)
            if len(value) == 6:
                return str(value[0]) + "." + str(value[1]) + "." + str(value[2]) + "." + str(value[3]) + "." + str(value[4]) + "." + str(value[5])
            raise ValueError("Invalid Logical name.")
        return str(value)

    @classmethod
    def logicalNameToBytes(cls, value):
        if not value:
            return bytearray(6)
        items = value.split('.')
        if len(items) != 6:
            raise ValueError("Invalid Logical name.")
        buff = bytearray(6)
        pos = 0
        for it in items:
            v = int(it)
            if v < 0 or v > 255:
                raise ValueError("Invalid Logical name.")
            buff[pos] = int(v)
            pos += 1
        return buff

    @classmethod
    def getGeneralizedTime(cls, dateString):
        year = int(dateString[0:4])
        month = int(dateString[4:6])
        day = int(dateString[6:8])
        hour = int(dateString[8:10])
        minute = int(dateString[10:12])
        #If UTC time.
        if dateString.endsWith("Z"):
            if len(dateString) > 13:
                second = int(dateString[12:14])
            return datetime(year, month, day, hour, minute, second, 0, tzinfo=tzoffset(None, 0))

        if len(dateString) > 17:
            second = int(dateString.substring(12, 14))
        tz = dateString[dateString.length() - 4:]
        return datetime(year, month, day, hour, minute, second, 0, tzinfo=tzoffset(None, tz))

    @classmethod
    def generalizedTime(cls, value):
        #Convert to UTC time.
        value = value.astimezone(pytz.utc)
        sb = cls.integerString(value.year, 4)
        sb += cls.integerString(value.month, 2)
        sb += cls.integerString(value.day, 2)
        sb += cls.integerString(value.hour, 2)
        sb += cls.integerString(value.minute, 2)
        sb += cls.integerString(value.second, 2)
        #UTC time.
        sb += "Z"
        return sb

    @classmethod
    def encryptManufacturer(cls, flagName):
        if len(flagName) != 3:
            raise ValueError("Invalid Flag name.")
        value = ((flagName.charAt(0) - 0x40) & 0x1f)
        value <<= 5
        value += ((flagName.charAt(0) - 0x40) & 0x1f)
        value <<= 5
        value += ((flagName.charAt(0) - 0x40) & 0x1f)
        return value

    @classmethod
    def decryptManufacturer(cls, value):
        tmp = (value >> 8 | value << 8)
        c = str(((tmp & 0x1f) + 0x40))
        tmp = (tmp >> 5)
        c1 = str(((tmp & 0x1f) + 0x40))
        tmp = (tmp >> 5)
        c2 = str(((tmp & 0x1f) + 0x40))
        return str(c2, c1, c)

    @classmethod
    def idisSystemTitleToString(cls, st):
        sb = '\r'
        sb += "IDIS system title:\r"
        sb += "Manufacturer Code: "
        sb += cls.__getChar(st[0]) + cls.__getChar(st[1]) + cls.__getChar(st[2])
        sb += "\rFunction type: "
        ft = st[4] >> 4
        add = False
        if (ft & 0x1) != 0:
            sb += "Disconnector extension"
            add = True
        if (ft & 0x2) != 0:
            if add:
                sb += ", "
            add = True
            sb += "Load Management extension"

        if (ft & 0x4) != 0:
            if add:
                sb += ", "
            sb += "Multi Utility extension"
        #Serial number
        sn = (st[4] & 0xF) << 24
        sn |= st[5] << 16
        sn |= st[6] << 8
        sn |= st[7]
        sb += '\r'
        sb += "Serial number: "
        sb += str(sn) + '\r'
        return sb

    @classmethod
    def dlmsSystemTitleToString(cls, st):
        sb = '\r'
        sb += "IDIS system title:\r"
        sb += "Manufacturer Code: "
        sb += cls.__getChar(st[0]) + cls.__getChar(st[1]) + cls.__getChar(st[2])
        sb += "Serial number: "
        sb += cls.__getChar(st[3]) + cls.__getChar(st[4]) + cls.__getChar(st[5]) + cls.__getChar(st[6]) + cls.__getChar(st[7])
        return sb

    @classmethod
    def uniSystemTitleToString(cls, st):
        sb = '\r'
        sb += "UNI/TS system title:\r"
        sb += "Manufacturer: "
        m = st[0] << 8 | st[1]
        sb += cls.decryptManufacturer(m)
        sb += "\rSerial number: "
        sb += GXByteBuffer.toHex((st[7], st[6], st[5], st[4], st[3], st[2]), False)
        return sb

    @classmethod
    def __getChar(cls, ch):
        try:
            return str(chr(ch))
        except Exception:
            #If python 2.7 is used.
            #pylint: disable=undefined-variable
            return str(unichr(ch))

    @classmethod
    def systemTitleToString(cls, standard, st):
        ###Conver system title to string.
        #pylint: disable=too-many-boolean-expressions
        if standard == Standard.ITALY or not cls.__getChar(st[0]).isalpha() or \
            not cls.__getChar(st[1]).isalpha() or not cls.__getChar(st[2]).isalpha():
            return cls.uniSystemTitleToString(st)
        if standard == Standard.IDIS or not cls.__getChar(st[3]).isdigit() or \
            not cls.__getChar(st[4]).isdigit() or not cls.__getChar(st[5]).isdigit() or \
            not cls.__getChar(st[6]).isdigit() or not cls.__getChar(st[7]).isdigit():
            return cls.idisSystemTitleToString(st)
        return cls.dlmsSystemTitleToString(st)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from .enums import ControlState, ControlMode

# pylint: disable=too-many-instance-attributes
class GXDLMSDisconnectControl(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSDisconnectControl
    """

    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.DISCONNECT_CONTROL, ln, sn)
        self.controlState = ControlState.DISCONNECTED
        self.controlMode = ControlMode.NONE
        self.outputState = False

    #
    # Forces the disconnect control object into 'disconnected' state if
    #      remote
    # disconnection is enabled.
    #
    def remoteDisconnect(self, client):
        return client.method(self, 1, int(0), DataType.INT8)

    #
    # Forces the disconnect control object into the
    #      'ready_for_reconnection'
    # state if a direct remote reconnection is disabled.
    #
    def remoteReconnect(self, client):
        return client.method(self, 2, int(0), DataType.INT8)

    def getValues(self):
        return [self.logicalName,
                self.outputState,
                self.controlState,
                self.controlMode]

    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  OutputState
        if all_ or self.canRead(2):
            attributes.append(2)
        #  ControlState
        if all_ or self.canRead(3):
            attributes.append(3)
        #  ControlMode
        if all_ or self.canRead(4):
            attributes.append(4)
        return attributes

    def getAttributeCount(self):
        return 4

    def getMethodCount(self):
        return 2

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.BOOLEAN
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.ENUM
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.outputState
        if e.index == 3:
            return self.controlState
        if e.index == 4:
            return self.controlMode
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if e.value is None:
                self.outputState = False
            else:
                self.outputState = e.value
        elif e.index == 3:
            #pylint: disable=bad-option-value,redefined-variable-type
            if e.value is None:
                self.controlState = ControlState.DISCONNECTED
            else:
                self.controlState = ControlState(e.value)
        elif e.index == 4:
            #pylint: disable=bad-option-value,redefined-variable-type
            if e.value is None:
                self.controlMode = ControlMode.NONE
            else:
                self.controlMode = ControlMode(e.value)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.outputState = reader.readElementContentAsInt("OutputState") != 0
        self.controlState = ControlState(reader.readElementContentAsInt("ControlState"))
        self.controlMode = ControlMode(reader.readElementContentAsInt("ControlMode"))

    def save(self, writer):
        writer.writeElementString("OutputState", self.outputState)
        if self.controlState:
            writer.writeElementString("ControlState", self.controlState)
        if self.controlMode:
            writer.writeElementString("ControlMode", self.controlMode)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from .GXDLMSScheduleEntry import GXDLMSScheduleEntry
from ..GXDateTime import GXDateTime

# pylint: disable=too-many-instance-attributes
class GXDLMSSchedule(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSSchedule
    """
    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        super(GXDLMSSchedule, self).__init__(ObjectType.SCHEDULE, ln, sn)
        # Specifies the scripts to be executed at given times.
        self.entries = list()

    def getValues(self):
        return [self.logicalName,
                self.entries]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Entries
        if all_ or not self.isRead(2):
            attributes.append(2)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):

        return 2

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):

        return 3

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.entries = []
            for it in e.value:
                item = GXDLMSScheduleEntry()
                item.index = it[0]
                item.enable = it[1]
                item.logicalName = _GXCommon.changeType(it[2], DataType.OCTET_STRING)
                item.scriptSelector = it[3]
                item.switchTime = _GXCommon.changeType(it[4], DataType.DATETIME)
                item.validityWindow = it[5]
                item.execWeekdays = it[6]
                item.execSpecDays = it[7]
                item.beginDate = _GXCommon.changeType(it[8], DataType.DATETIME)
                item.endDate = _GXCommon.changeType(it[9], DataType.DATETIME)
                self.entries.append(item)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.entries = []
        if reader.isStartElement("Entries", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSScheduleEntry()
                it.index = reader.readElementContentAsInt("Index")
                it.enable = reader.readElementContentAsInt("Enable") != 0
                it.logicalName = reader.readElementContentAsString("LogicalName")
                it.scriptSelector = reader.readElementContentAsInt("ScriptSelector")
                it.switchTime = reader.readElementContentAsObject("SwitchTime", GXDateTime())
                it.validityWindow = reader.readElementContentAsInt("ValidityWindow")
                it.execWeekdays = reader.readElementContentAsString("ExecWeekdays")
                it.execSpecDays = reader.readElementContentAsString("ExecSpecDays")
                it.beginDate = reader.readElementContentAsObject("BeginDate", GXDateTime())
                it.endDate = reader.readElementContentAsObject("EndDate", GXDateTime())
                self.entries.append(it)
            reader.readEndElement("Entries")

    def save(self, writer):
        if self.entries:
            writer.writeStartElement("Entries")
            for it in self.entries:
                writer.writeStartElement("Item")
                writer.writeElementString("Index", it.index)
                writer.writeElementString("Enable", it.enable)
                writer.writeElementString("LogicalName", it.logicalName)
                writer.writeElementString("ScriptSelector", it.scriptSelector)
                writer.writeElementObject("SwitchTime", it.switchTime)
                writer.writeElementString("ValidityWindow", it.validityWindow)
                writer.writeElementString("ExecWeekdays", it.execWeekdays)
                writer.writeElementString("ExecSpecDays", it.execSpecDays)
                writer.writeElementObject("BeginDate", it.beginDate)
                writer.writeElementObject("EndDate", it.endDate)
                writer.writeEndElement()
            writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .TranslatorOutputType import TranslatorOutputType
from .TranslatorTags import TranslatorTags
from .ConfirmedServiceError import ConfirmedServiceError
from .ServiceError import ServiceError
from .enums import Command, ErrorCode, DataType, AccessMode, Service, ObjectType, MethodAccessMode
from .VariableAccessSpecification import VariableAccessSpecification
from .internal._GXCommon import _GXCommon
from .internal._GXDataInfo import _GXDataInfo
from .GXByteBuffer import GXByteBuffer
from .GXDLMS import GXDLMS
from .GXDLMSSNParameters import GXDLMSSNParameters
from .SingleReadResponse import SingleReadResponse
from .GXSNInfo import GXSNInfo
from .ValueEventArgs import ValueEventArgs
from .GXDLMSLongTransaction import GXDLMSLongTransaction
from .SingleWriteResponse import SingleWriteResponse

# pylint: disable=bad-option-value,too-many-locals,too-many-arguments,old-style-class
class GXDLMSSNCommandHandler:
    #Constructor.
    def __init__(self):
        pass

    @classmethod
    def handleRead(cls, settings, server, type_, data, list_, reads, actions, replyData, xml):
        #  GetRequest normal
        sn = data.getInt16()
        if xml:
            if xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendStartTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
            else:
                sn &= 0xFFFF
            if type_ == VariableAccessSpecification.PARAMETERISED_ACCESS:
                xml.appendStartTag(Command.READ_REQUEST, VariableAccessSpecification.PARAMETERISED_ACCESS)
                xml.appendLine(Command.READ_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME, "Value", xml.integerToHex(sn, 4))
                xml.appendLine(TranslatorTags.SELECTOR, "Value", xml.integerToHex(data.getUInt8(), 2))
                di = _GXDataInfo()
                di.xml = xml
                xml.appendStartTag(TranslatorTags.PARAMETER)
                _GXCommon.getData(data, di)
                xml.appendEndTag(TranslatorTags.PARAMETER)
                xml.appendEndTag(Command.READ_REQUEST, VariableAccessSpecification.PARAMETERISED_ACCESS)
            else:
                xml.appendLine(Command.READ_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME, "Value", xml.integerToHex(sn, 4))
            if xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendEndTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
            return
        sn = sn & 0xFFFF
        i = cls.findSNObject(server, server.settings, sn)
        e = ValueEventArgs(server, i.item, i.index, 0, None)
        e.action = i.action
        if type_ == VariableAccessSpecification.PARAMETERISED_ACCESS:
            e.selector = data.getUInt8()
            di = _GXDataInfo()
            e.parameters = _GXCommon.getData(data, di)
        #  Return error if connection is not established.
        if not settings.acceptConnection() and (not e.action or e.target.shortName != 0xFA00 or e.index != 8):
            replyData.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        list_.append(e)
        if not e.action and server.onGetAttributeAccess(e) == AccessMode.NO_ACCESS:
            e.error = ErrorCode.READ_WRITE_DENIED
        elif e.action and server.onGetMethodAccess(e) == MethodAccessMode.NO_ACCESS:
            e.error = ErrorCode.READ_WRITE_DENIED
        else:
            if e.action:
                actions.append(e)
            else:
                reads.append(e)

    #
    # Handle read Block in blocks.
    #
    # @param data
    #            Received data.
    #
    @classmethod
    def handleReadBlockNumberAccess(cls, settings, server, data, replyData, xml):
        blockNumber = data.getUInt16()
        if xml:
            xml.appendStartTag(Command.READ_REQUEST, VariableAccessSpecification.BLOCK_NUMBER_ACCESS)
            xml.appendLine(TranslatorTags.BLOCK_NUMBER, "Value", xml.integerToHex(blockNumber, 4))
            xml.appendEndTag(Command.READ_REQUEST, VariableAccessSpecification.BLOCK_NUMBER_ACCESS)
            return
        bb = GXByteBuffer()
        if blockNumber != settings.blockIndex:
            bb.setUInt8(ErrorCode.DATA_BLOCK_NUMBER_INVALID)
            GXDLMS.getSNPdu(GXDLMSSNParameters(settings, Command.READ_RESPONSE, 1, SingleReadResponse.DATA_ACCESS_ERROR, bb, None), replyData)
            settings.resetBlockIndex()
            return
        if settings.index != settings.count and server.transaction.data.size() < settings.maxPduSize:
            reads = list()
            actions = list()
            for it in server.transaction.targets:
                if it.action:
                    actions.append(it)
                else:
                    reads.append(it)
            if reads:
                server.onPreRead(reads)
            if actions:
                server.onPreAction(actions)
            cls.getReadData(settings, server.transaction.targets, server.transaction.data)
            if reads:
                server.onPostRead(reads)
            if actions:
                server.onPostAction(actions)
        settings.increaseBlockIndex()
        p = GXDLMSSNParameters(settings, Command.READ_RESPONSE, 1, SingleReadResponse.DATA_BLOCK_RESULT, bb, server.transaction.data)
        p.multipleBlocks = True
        GXDLMS.getSNPdu(p, replyData)
        if server.transaction.data.size() == server.transaction.data.position:
            server.transaction = None
            settings.resetBlockIndex()
        else:
            server.transaction.data.strip()

    @classmethod
    def getReadData(cls, settings, list_, data):
        value = None
        first = True
        type_ = SingleReadResponse.DATA
        for e in list_:
            if e.handled:
                value = e.value
            else:
                if e.action:
                    value = e.target.invoke(settings, e)
                else:
                    value = e.target.getValue(settings, e)
            if e.error == ErrorCode.OK:
                if not first and list_:
                    data.setUInt8(SingleReadResponse.DATA)
                if e.action:
                    _GXCommon.setData(data, _GXCommon.getDLMSDataType(value), value)
                else:
                    GXDLMS.appendData(e.target, e.index, data, value)
            else:
                if not first and list_:
                    data.setUInt8(SingleReadResponse.DATA_ACCESS_ERROR)
                data.setUInt8(e.error)
                type_ = SingleReadResponse.DATA_ACCESS_ERROR
            first = False
        return type_

    @classmethod
    def handleReadDataBlockAccess(cls, settings, server, command, data, cnt, replyData, xml):
        bb = GXByteBuffer()
        lastBlock = data.getUInt8()
        blockNumber = data.getUInt16()
        if xml:
            if command == Command.WRITE_RESPONSE:
                xml.appendStartTag(TranslatorTags.WRITE_DATA_BLOCK_ACCESS)
            else:
                xml.appendStartTag(TranslatorTags.READ_DATA_BLOCK_ACCESS)
            xml.appendLine("<LastBlock Value=\"" + xml.integerToHex(lastBlock, 2) + "\" />")
            xml.appendLine("<BlockNumber Value=\"" + xml.integerToHex(blockNumber, 4) + "\" />")
            if command == Command.WRITE_RESPONSE:
                xml.appendEndTag(TranslatorTags.WRITE_DATA_BLOCK_ACCESS)
            else:
                xml.appendEndTag(TranslatorTags.READ_DATA_BLOCK_ACCESS)
            return
        if blockNumber != settings.blockIndex:
            bb.setUInt8(ErrorCode.DATA_BLOCK_NUMBER_INVALID)
            GXDLMS.getSNPdu(GXDLMSSNParameters(settings, command, 1, SingleReadResponse.DATA_ACCESS_ERROR, bb, None), replyData)
            settings.resetBlockIndex()
            return
        count = 1
        type1 = DataType.OCTET_STRING
        if command == Command.WRITE_RESPONSE:
            count = data.getUInt8()
            type1 = data.getUInt8()
        size = _GXCommon.getObjectCount(data)
        realSize = len(data) - data.position
        if count != 1 or type1 != DataType.OCTET_STRING or size != realSize:
            bb.setUInt8(ErrorCode.DATA_BLOCK_UNAVAILABLE)
            GXDLMS.getSNPdu(GXDLMSSNParameters(settings, command, cnt, SingleReadResponse.DATA_ACCESS_ERROR, bb, None), replyData)
            settings.resetBlockIndex()
            return
        if server.transaction is None:
            server.setTransaction(GXDLMSLongTransaction(None, command, data))
        else:
            server.transaction.data.set(data)
        if lastBlock == 0:
            #pylint: disable=bad-option-value,redefined-variable-type
            bb.setUInt16(blockNumber)
            settings.increaseBlockIndex()
            if command == Command.READ_RESPONSE:
                type2 = SingleReadResponse.BLOCK_NUMBER
            else:
                type2 = SingleWriteResponse.BLOCK_NUMBER
            GXDLMS.getSNPdu(GXDLMSSNParameters(settings, command, cnt, type2, None, bb), replyData)
            return
        if server.transaction:
            data.size(0)
            data.set(server.transaction.data)
            server.setTransaction(None)
        if command == Command.READ_RESPONSE:
            cls.handleReadRequest(settings, server, data, replyData, xml)
        else:
            cls.handleWriteRequest(settings, server, data, replyData, xml)
        settings.resetBlockIndex()

    @classmethod
    def handleReadRequest(cls, settings, server, data, replyData, xml):
        bb = GXByteBuffer()
        cnt = 0xFF
        list_ = list()
        if xml is None and not data:
            if server.transaction:
                return
            bb.set(replyData)
            replyData.clear()
            for it in server.transaction.targets:
                list_.append(it)
        else:
            cnt = _GXCommon.getObjectCount(data)
            reads = list()
            actions = list()
            if xml:
                xml.appendStartTag(Command.READ_REQUEST, "Qty", xml.integerToHex(cnt, 2))
            pos = 0
            while pos != cnt:
                type_ = data.getUInt8()
                if type_ in (VariableAccessSpecification.VARIABLE_NAME, VariableAccessSpecification.PARAMETERISED_ACCESS):
                    cls.handleRead(settings, server, type_, data, list_, reads, actions, replyData, xml)
                elif type_ == VariableAccessSpecification.BLOCK_NUMBER_ACCESS:
                    cls.handleReadBlockNumberAccess(settings, server, data, replyData, xml)
                    if xml:
                        xml.appendEndTag(Command.READ_REQUEST)
                    return
                elif type_ == VariableAccessSpecification.READ_DATA_BLOCK_ACCESS:
                    cls.handleReadDataBlockAccess(settings, server, Command.READ_RESPONSE, data, cnt, replyData, xml)
                    if xml:
                        xml.appendEndTag(Command.READ_REQUEST)
                    return
                else:
                    cls.returnSNError(settings, Command.READ_RESPONSE, ErrorCode.READ_WRITE_DENIED, replyData)
                    return
                pos += 1
            if reads:
                server.onPreRead(reads)
            if actions:
                server.onPreAction(actions)
        if xml:
            xml.appendEndTag(Command.READ_REQUEST)
            return
        requestType = cls.getReadData(settings, list_, bb)
        p = GXDLMSSNParameters(settings, Command.READ_RESPONSE, cnt, requestType, None, bb)
        GXDLMS.getSNPdu(p, replyData)
        if server.transaction() is None and (len(bb) != bb.position or settings.count != settings.index):
            reads = list()
            for it in list_:
                reads.append(it)
            if reads:
                server.onPostRead(reads)
            server.setTransaction(GXDLMSLongTransaction(reads, Command.READ_REQUEST, bb))
        elif server.transaction:
            replyData.set(bb)
            return

    @classmethod
    def returnSNError(cls, settings, cmd, error, replyData):
        bb = GXByteBuffer()
        bb.setUInt8(error)
        GXDLMS.getSNPdu(GXDLMSSNParameters(settings, cmd, 1, SingleReadResponse.DATA_ACCESS_ERROR, bb, None), replyData)
        settings.resetBlockIndex()

    @classmethod
    def findSNObject(cls, server, settings, sn):
        i = GXSNInfo()
        offset = [0]
        count = [0]
        for it in settings.objects:
            if sn >= it.shortName:
                if sn < it.shortName + it.getAttributeCount() * 8:
                    i.action = False
                    i.item = it
                    i.index = ((sn - it.shortName) / 8) + 1
                    break
                else:
                    GXDLMS.getActionInfo(it.objectType, offset, count)
                    if sn < it.getShortName() + offset[0] + (8 * count[0]):
                        i.item = it
                        i.action = True
                        i.index = (sn - it.shortName - offset[0]) / 8 + 1
                        break
        if i.item is None and server:
            i.item = server.onFindObject(ObjectType.NONE, sn, None)
        return i

    @classmethod
    def handleWriteRequest(cls, settings, server, data, replyData, xml):
        if xml is None and not settings.acceptConnection:
            replyData.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        type_ = int()
        value = None
        targets = list()
        cnt = _GXCommon.getObjectCount(data)
        if xml:
            xml.appendStartTag(Command.WRITE_REQUEST)
            xml.appendStartTag(TranslatorTags.LIST_OF_VARIABLE_ACCESS_SPECIFICATION, "Qty", xml.integerToHex(cnt, 2))
            if xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendStartTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
        results = GXByteBuffer(cnt)
        pos = 0
        while pos != cnt:
            type_ = data.getUInt8()
            if type_ == VariableAccessSpecification.VARIABLE_NAME:
                sn = data.getUInt16()
                if xml:
                    xml.appendLine(Command.WRITE_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME, "Value", xml.integerToHex(sn, 4))
                else:
                    i = cls.findSNObject(server, server.settings, sn)
                    targets.append(i)
                    if i is None:
                        results.setUInt8(ErrorCode.UNDEFINED_OBJECT)
                    else:
                        results.setUInt8(ErrorCode.OK)
            elif type_ == VariableAccessSpecification.WRITE_DATA_BLOCK_ACCESS:
                cls.handleReadDataBlockAccess(settings, server, Command.WRITE_RESPONSE, data, cnt, replyData, xml)
                if xml is None:
                    return
            else:
                results.setUInt8(ErrorCode.HARDWARE_FAULT)
            pos += 1
        if xml:
            if xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendEndTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
            xml.appendEndTag(TranslatorTags.LIST_OF_VARIABLE_ACCESS_SPECIFICATION)
        cnt = _GXCommon.getObjectCount(data)
        di = _GXDataInfo()
        if xml:
            di.xml = xml
            xml.appendStartTag(TranslatorTags.LIST_OF_DATA, "Qty", xml.integerToHex(cnt, 2))
        pos = 0
        while pos != cnt:
            di.clear()
            if xml:
                if xml.outputType == TranslatorOutputType.STANDARD_XML:
                    xml.appendStartTag(Command.WRITE_REQUEST << 8 | SingleReadResponse.DATA)
                value = _GXCommon.getData(data, di)
                if not di.complete:
                    value = GXByteBuffer.hex(data.data, False, data.position, len(data) - data.position)
                    xml.appendLine(_GXCommon.DATA_TYPE_OFFSET + di.type, "Value", str(value))
                if xml.outputType == TranslatorOutputType.STANDARD_XML:
                    xml.appendEndTag(Command.WRITE_REQUEST << 8 | SingleReadResponse.DATA)
            elif results.getUInt8(pos) == 0:
                target = targets[pos]
                value = _GXCommon.getData(data, di)
                if isinstance(value, bytearray):
                    dt = target.getItem().getDataType(target.index)
                    if dt not in (DataType.NONE, DataType.OCTET_STRING):
                        value = _GXCommon.changeType(value, dt)
                e = ValueEventArgs(server, target.getItem(), target.index, 0, None)
                am = server.onGetAttributeAccess(e)
                if am not in (AccessMode.WRITE, AccessMode.READ_WRITE):
                    results.setUInt8(pos, ErrorCode.READ_WRITE_DENIED)
                else:
                    e.value = value
                    server.onPreWrite(list(e))
                    if e.error != ErrorCode.OK:
                        results.setUInt8(pos, e.error)
                    elif not e.handled:
                        target.item.setValue(settings, e)
                    server.onPostWrite((e))
            pos += 1
        if xml:
            xml.appendEndTag(TranslatorTags.LIST_OF_DATA)
            xml.appendEndTag(Command.WRITE_REQUEST)
            return
        bb = GXByteBuffer((2 * cnt))
        ret = int()
        pos = 0
        while pos != cnt:
            ret = results.getUInt8(pos)
            if ret != 0:
                bb.setUInt8(1)
            bb.setUInt8(ret)
            pos += 1
        p = GXDLMSSNParameters(settings, Command.WRITE_RESPONSE, cnt, 0xFF, None, bb)
        GXDLMS.getSNPdu(p, replyData)

    @classmethod
    def handleInformationReport(cls, settings, reply, list_):
        data = reply.data
        reply.time = None
        len_ = data.getUInt8()
        tmp = None
        if len_ != 0:
            tmp = bytearray(len_)
            data.get(tmp)
            reply.t = _GXCommon.changeType(tmp, DataType.DATETIME)
        type_ = 0
        ot = TranslatorOutputType.SIMPLE_XML
        if reply.xml:
            ot = reply.xml.outputType
        count = _GXCommon.getObjectCount(reply.data)
        if reply.xml:
            reply.xml.appendStartTag(Command.INFORMATION_REPORT)
            if reply.time:
                reply.xml.appendComment(str(reply.time))
                if ot == TranslatorOutputType.SIMPLE_XML:
                    reply.xml.appendLine(TranslatorTags.CURRENT_TIME, None, GXByteBuffer.hex(tmp, False))
                else:
                    reply.xml.appendLine(TranslatorTags.CURRENT_TIME, None, _GXCommon.generalizedTime(reply.time))
            reply.xml.appendStartTag(TranslatorTags.LIST_OF_VARIABLE_ACCESS_SPECIFICATION, "Qty", reply.xml.integerToHex(count, 2))
        pos = 0
        while pos != count:
            type_ = data.getUInt8()
            if type_ == VariableAccessSpecification.VARIABLE_NAME:
                sn = data.getUInt16()
                if reply.xml:
                    if ot == TranslatorOutputType.STANDARD_XML:
                        reply.xml.appendStartTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
                    reply.xml.appendLine(Command.WRITE_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME, "Value", reply.xml.integerToHex(sn, 4))
                    if ot == TranslatorOutputType.STANDARD_XML:
                        reply.xml.appendEndTag(TranslatorTags.VARIABLE_ACCESS_SPECIFICATION)
                else:
                    info = cls.findSNObject(None, settings, sn)
                    if info.item:
                        list_.append((info.item, info.index))
                    else:
                        print("InformationReport message. Unknown object : " + str(sn))
            pos += 1
        if reply.xml:
            reply.xml.appendEndTag(TranslatorTags.LIST_OF_VARIABLE_ACCESS_SPECIFICATION)
            reply.xml.appendStartTag(TranslatorTags.LIST_OF_DATA, "Qty", reply.xml.integerToHex(count, 2))
        count = _GXCommon.getObjectCount(reply.data)
        di = _GXDataInfo()
        di.xml = (reply.xml)
        pos = 0
        while pos != count:
            di.clear()
            if reply.xml:
                if ot == TranslatorOutputType.STANDARD_XML:
                    reply.xml.appendStartTag(Command.WRITE_REQUEST << 8 | SingleReadResponse.DATA)
                _GXCommon.getData(reply.data, di)
                if ot == TranslatorOutputType.STANDARD_XML:
                    reply.xml.appendEndTag(Command.WRITE_REQUEST << 8 | SingleReadResponse.DATA)
            else:
                v = ValueEventArgs(list_[pos].key, list_[pos].value, 0, None)
                v.value = _GXCommon.getData(reply.data, di)
                list_.get(pos).getKey().setValue(settings, v)
            pos += 1
        if reply.xml:
            reply.xml.appendEndTag(TranslatorTags.LIST_OF_DATA)
            reply.xml.appendEndTag(Command.INFORMATION_REPORT)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gurux_dlms",
    version="1.0.42",
    author="Gurux Ltd",
    author_email="gurux@gurux.org",
    description="Gurux DLMS library for Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gurux/gurux.dlms.python",
    packages=setuptools.find_packages(),
    package_data={'gurux_dlms': ['OBISCodes.txt', 'India.txt', 'Italy.txt', 'SaudiArabia.txt']},
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    install_requires=['python-dateutil', 'pytz'],
)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.Security import Security
from .GXByteBuffer import GXByteBuffer
from .CountType import CountType
from .objects.enums.SecuritySuite import SecuritySuite
from .GXDLMSChipperingStream import GXDLMSChipperingStream
from .internal._GXCommon import _GXCommon
from .enums.Command import Command

#pylint: disable=too-many-instance-attributes,too-many-public-methods
class GXDLMSChippering:

    #
    #      * Get nonse from frame counter and system title.
    #      * @param invocationCounter Invocation counter.
    #      * @param systemTitle System title.
    #      * @return Generated nonse.
    #
    @classmethod
    def getNonse(cls, invocationCounter, systemTitle):
        nonce = bytearray(12)
        nonce[0:7] = systemTitle
        nonce[8] = ((invocationCounter >> 24) & 0xFF)
        nonce[9] = ((invocationCounter >> 16) & 0xFF)
        nonce[10] = ((invocationCounter >> 8) & 0xFF)
        nonce[11] = (invocationCounter & 0xFF)
        return nonce

    @classmethod
    def encryptAesGcm(cls, p, plainText):
        p.countTag = None
        data = GXByteBuffer()
        if p.type_ == CountType.PACKET:
            data.setUInt8(p.security)
        tmp = bytearray(4)
        invocationCounter = 0
        if p.securitySuite == SecuritySuite.AES_GCM_128:
            invocationCounter = p.invocationCounter
        tmp[0] = ((invocationCounter >> 24) & 0xFF)
        tmp[1] = ((invocationCounter >> 16) & 0xFF)
        tmp[2] = ((invocationCounter >> 8) & 0xFF)
        tmp[3] = (invocationCounter & 0xFF)
        aad = cls.getAuthenticatedData(p, plainText)
        iv = cls.getNonse(invocationCounter, p.systemTitle)
        gcm = GXDLMSChipperingStream(p.security, True, p.blockCipherKey, aad, iv, None)
        #  Encrypt the secret message
        if p.security != Security.AUTHENTICATION:
            gcm.write(plainText)
        ciphertext = gcm.flushFinalBlock()
        if p.security == Security.AUTHENTICATION:
            if p.type_ == CountType.PACKET:
                data.set(tmp)
            if (p.type_ & CountType.DATA) != 0:
                data.set(plainText)
            if (p.type_ & CountType.TAG) != 0:
                p.countTag = gcm.tag
                data.set(p.countTag)
        elif p.security == Security.ENCRYPTION:
            if p.type_ == CountType.PACKET:
                data.set(tmp)
            data.set(ciphertext)
        elif p.security == Security.AUTHENTICATION_ENCRYPTION:
            if p.type_ == CountType.PACKET:
                data.set(tmp)
            if (p.type_ & CountType.DATA) != 0:
                data.set(ciphertext)
            if (p.type_ & CountType.TAG) != 0:
                p.countTag = gcm.tag
                data.set(p.countTag)
        else:
            raise ValueError("security")
        if p.type_ == CountType.PACKET:
            tmp2 = GXByteBuffer(10 + len(data))
            tmp2.setUInt8(p.tag)
            _GXCommon.setObjectCount(len(data), tmp2)
            tmp2.set(data, 0, len(data))
            data = tmp2
        crypted = data.array()
        return crypted

    @classmethod
    def getAuthenticatedData(cls, p, plainText):
        data = GXByteBuffer()
        sc = p.security | p.securitySuite
        if p.security == Security.AUTHENTICATION:
            data.setUInt8(sc)
            data.set(p.authenticationKey)
            data.set(plainText)
        elif p.security == Security.AUTHENTICATION_ENCRYPTION:
            data.setUInt8(sc)
            data.set(p.authenticationKey)
            if p.securitySuite != SecuritySuite.AES_GCM_128:
                #  transaction-id
                transactionId = GXByteBuffer()
                transactionId.setUInt64(p.invocationCounter)
                data.setUInt8(8)
                data.set(transactionId)
                #  originator-system-title
                _GXCommon.setObjectCount(len(p.systemTitle), data)
                data.set(p.systemTitle)
                #  recipient-system-title
                _GXCommon.setObjectCount(len(p.recipientSystemTitle), data)
                data.set(p.recipientSystemTitle)
                #  date-time not present
                data.setUInt8(0)
                #  other-information not present
                data.setUInt8(0)
        elif p.security == Security.ENCRYPTION:
            data.set(p.authenticationKey)
        return data.array()

    #
    #      * Decrypt data.
    #      *
    #      * @param c
    #      * Cipher settings.
    #      * @param p
    #      * GMAC Parameter.
    #      * @return Encrypted data.
    #
    @classmethod
    def decryptAesGcm(cls, p, data):
        # pylint: disable=too-many-locals
        if not data or len(data) - data.position < 2:
            raise ValueError("cryptedData")
        tmp = []
        len_ = 0
        cmd = data.getUInt8()
        if cmd in (Command.GENERAL_GLO_CIPHERING, Command.GENERAL_DED_CIPHERING):
            len_ = _GXCommon.getObjectCount(data)
            title = bytearray(len_)
            data.get(title)
            p.systemTitle = title
            if p.xml and p.xml.comments:
                p.xml.appendComment(_GXCommon.systemTitleToString(0, p.systemTitle))
        elif cmd in (Command.GENERAL_CIPHERING, Command.GLO_INITIATE_REQUEST, Command.GLO_INITIATE_RESPONSE, Command.GLO_READ_REQUEST, Command.GLO_READ_RESPONSE, \
            Command.GLO_WRITE_REQUEST, Command.GLO_WRITE_RESPONSE, Command.GLO_GET_REQUEST, Command.GLO_GET_RESPONSE, Command.GLO_SET_REQUEST, \
            Command.GLO_SET_RESPONSE, Command.GLO_METHOD_REQUEST, Command.GLO_METHOD_RESPONSE, Command.GLO_EVENT_NOTIFICATION_REQUEST,\
            Command.DED_GET_REQUEST, Command.DED_GET_RESPONSE, Command.DED_SET_REQUEST, Command.DED_SET_RESPONSE, Command.DED_METHOD_REQUEST,\
            Command.DED_METHOD_RESPONSE, Command.DED_EVENT_NOTIFICATION):
            pass
        else:
            raise ValueError("cryptedData")
        value = 0
        transactionId = 0
        if cmd == Command.GENERAL_CIPHERING:
            len_ = _GXCommon.getObjectCount(data)
            tmp = bytearray(len_)
            data.get(tmp)
            t = GXByteBuffer(tmp)
            transactionId = t.getInt64()
            len_ = _GXCommon.getObjectCount(data)
            tmp = bytearray(len_)
            data.get(tmp)
            p.setSystemTitle(tmp)
            len_ = _GXCommon.getObjectCount(data)
            tmp = bytearray(len_)
            data.get(tmp)
            p.setRecipientSystemTitle(tmp)
            #  Get date time.
            len_ = _GXCommon.getObjectCount(data)
            if len_ != 0:
                tmp = bytearray(len_)
                data.get(tmp)
                p.dateTime = tmp
            #  other-information
            len_ = data.getUInt8()
            if len_ != 0:
                tmp = bytearray(len_)
                data.get(tmp)
                p.otherInformation = tmp
            #  KeyInfo OPTIONAL
            len_ = data.getUInt8()
            #  AgreedKey CHOICE tag.
            data.getUInt8()
            #  key-parameters
            len_ = data.getUInt8()
            value = data.getUInt8()
            p.setKeyParameters(value)
            if value == 1:
                #  KeyAgreement.ONE_PASS_DIFFIE_HELLMAN
                #  key-ciphered-data
                len_ = _GXCommon.getObjectCount(data)
                tmp = bytearray(len_)
                data.get(tmp)
                p.keyCipheredData = tmp
            elif value == 2:
                #  KeyAgreement.STATIC_UNIFIED_MODEL
                len_ = _GXCommon.getObjectCount(data)
                if len_ != 0:
                    raise ValueError("Invalid key parameters")
            else:
                raise ValueError("key-parameters")
        len_ = _GXCommon.getObjectCount(data)
        p.cipheredContent = data.remaining()
        sc = data.getUInt8()
        security = Security(sc & 0x30)
        ss = SecuritySuite(sc & 0x3)
        if ss != SecuritySuite.AES_GCM_128:
            raise ValueError("Decrypt failed. Invalid security suite.")
        p.security = security
        invocationCounter = data.getUInt32()
        p.invocationCounter = invocationCounter
        tag = bytearray(12)
        encryptedData = None
        if security == Security.AUTHENTICATION:
            len_ = len(data) - data.position() - 12
            encryptedData = bytearray(len_)
            data.get(encryptedData)
            data.get(tag)
            #  Check tag.
            cls.encryptAesGcm(p, encryptedData)
            if not GXDLMSChipperingStream.tagsEquals(tag, p.countTag):
                if transactionId != 0:
                    p.setInvocationCounter(transactionId)
                if not p.xml:
                    raise ValueError("Decrypt failed. Invalid tag.")
                p.xml.appendComment("Decrypt failed. Invalid tag.")
            return encryptedData
        ciphertext = None
        if security == Security.ENCRYPTION:
            len_ = len(data) - data.position
            ciphertext = bytearray(len_)
            data.get(ciphertext)
        elif security == Security.AUTHENTICATION_ENCRYPTION:
            len_ = len(data) - data.position - 12
            ciphertext = bytearray(len_)
            data.get(ciphertext)
            data.get(tag)
        aad = cls.getAuthenticatedData(p, ciphertext)
        iv = cls.getNonse(invocationCounter, p.systemTitle)
        gcm = GXDLMSChipperingStream(security, True, p.blockCipherKey, aad, iv, tag)
        gcm.write(ciphertext)
        if transactionId != 0:
            p.setInvocationCounter(transactionId)
        return gcm.flushFinalBlock()
